from typing import Union
from abc import ABCMeta, abstractmethod

from PIL import Image

from eyewitness.detection_utils import DetectionResult
from eyewitness.image_id import ImageId


class ObjectDetector(metaclass=ABCMeta):
    @abstractmethod
    def detect(self, image: Image, image_id: Union[str, ImageId]) -> DetectionResult:
        pass



import io

import arrow
from flask import Flask
from flask import request

from eyewitness.image_id import ImageId
from eyewitness.image_utils import ImageHandler


class ObjectDetectionFlaskWrapper(object):
    def __init__(self, obj_detector, detection_result_handler):
        app = Flask(__name__)
        self.app = app
        self.obj_detector = obj_detector
        self.detection_result_handler = detection_result_handler

        @app.route("/detect_image_byte", methods=['POST'])
        def detect_image_bytes_objs():
            if 'store_image_path' not in request.headers:
                channel = request.headers['channel']
                datetime = arrow.now()
                file_format = request.headers.get('file_format', 'jpg')
                image_id = ImageId(channel, datetime, format=file_format)
            else:
                image_id = request.headers['store_image_path']

            store_raw_image = request.headers.get('store_raw', True)

            # read data from Bytes
            data = request.data
            image_data_raw = io.BytesIO(bytearray(data))
            image_raw = ImageHandler.read_image_bytes(image_data_raw)

            if store_raw_image:
                ImageHandler.save(image_raw, image_id)

            # detect objs
            detection_result = self.obj_detector.detect(image_raw, image_id)
            detection_result_handler.handle(detection_result)
            return "successfully detected"

        @app.route("/detect_image_file", methods=['POST'])
        def detect_image_file_objs():
            image_id = request.headers['image_path']
            image_raw = ImageHandler.read_image_file(image_id)
            detection_result = self.obj_detector.detect(image_raw, image_id)
            detection_result_handler.handle(detection_result)
            return "successfully detected"



import os
from abc import ABCMeta, abstractmethod
from io import BytesIO
from typing import List, Optional, Union

import pkg_resources
from PIL import Image, ImageFont, ImageDraw
import numpy as np

from eyewitness.utils import make_path
from eyewitness.image_id import ImageId
from eyewitness.detection_utils import BoundedBoxObject


DEFAULT_FONT_PATH = pkg_resources.resource_filename('eyewitness', 'font/FiraMono-Medium.otf')


class ImageHandler(object):
    """
    common functions for image processing
    """
    @classmethod
    def save(cls, image: Image.Image, output_path: Union[str, ImageId]):
        if isinstance(output_path, ImageId):
            output_path = str(output_path)
        folder = os.path.dirname(output_path)
        make_path(folder)
        image.save(output_path)

    @classmethod
    def read_image_file(cls, image_path: str) -> Image.Image:
        """
        Image.open read from file
        """
        return Image.open(image_path)

    @classmethod
    def read_image_bytes(cls, image_byte: BytesIO) -> Image.Image:
        """
        Image.open support BytesIO input
        """
        return Image.open(image_byte)

    @classmethod
    def draw_bbox(
            cls,
            image: Image.Image,
            detections: List[BoundedBoxObject],
            colors: Optional[dict] = None,
            font_path: str = DEFAULT_FONT_PATH):
        if colors is None:
            colors = {}

        font = ImageFont.truetype(
            font=font_path,
            size=np.floor(3e-2 * image.size[1] + 0.5).astype('int32'))
        thickness = (image.size[0] + image.size[1]) // 300

        for (left, top, right, bottom, predicted_class, score, _) in detections:
            label = '{} {:.2f}'.format(predicted_class, score)
            draw = ImageDraw.Draw(image)
            label_size = draw.textsize(label, font)

            # creating bbox on images
            if top - label_size[1] >= 0:
                text_origin = np.array([left, top - label_size[1]])
            else:
                text_origin = np.array([left, top + 1])

            for i in range(thickness):
                draw.rectangle(
                    [left + i, top + i, right - i, bottom - i],
                    outline=colors.get(predicted_class, 'red'))
            draw.rectangle(
                [tuple(text_origin), tuple(text_origin + label_size)],
                fill=colors.get(predicted_class, 'red'))
            draw.text(text_origin, label, fill=(0, 0, 0), font=font)


class ImageProducer(metaclass=ABCMeta):

    def __init__(self, setting: dict):
        self.setting = setting

    @abstractmethod
    def generate_img(self) -> np.array:
        pass

    @abstractmethod
    def post_img(self, image: np.array, destination: str):
        pass


def swap_channel_rgb_bgr(image: np.array):
    """
    reverse the color channel image
    convert image (w, h, c) with channel rgb -> bgr
    convert image (w, h, c) with channel bgr -> rgb
    """
    image = image[:, :, ::-1]
    return image



import json
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from typing import List

from eyewitness.image_id import ImageId


BoundedBoxObject = namedtuple(
    'BoundedBoxObject', ['x1', 'y1', 'x2', 'y2', 'class_name', 'score', 'meta'])
Type_Serialization_Mapping = {'bbox': BoundedBoxObject}


class DetectionResult(object):
    """
    represent detection result of a image.
    """

    def __init__(self, image_dict, detection_type='bbox'):
        self.detection_type = Type_Serialization_Mapping[detection_type]
        if isinstance(image_dict['image_id'], ImageId):
            image_dict['image_id'] = str(image_dict['image_id'])
        self.image_dict = image_dict

    @property
    def image_id(self) -> str:
        return self.image_dict['image_id']

    @property
    def drawn_image_path(self) -> str:
        return self.image_dict.get('drawn_image_path', '')

    @property
    def detected_objects(self) -> List:
        """
        list of detected objects (will translate type into detection_type)
        """
        detected_objects = self.image_dict.get('detected_objects', [])
        return [self.detection_type(*i) for i in detected_objects]

    @classmethod
    def from_json(cls, json_str: str):
        img_dict = json.loads(json_str)
        return cls(img_dict)

    def to_json(self) -> str:
        json_str = json.dumps(self.image_dict)
        return json_str

    def to_dict(self) -> dict:
        return self.image_dict


class DetectionResultHandler(metaclass=ABCMeta):
    @abstractmethod
    def handle(self, detection_result: DetectionResult):
        pass



from arrow import Arrow


class ImageId(object):
    """
    the target of ImageId is used to standardize the image_id format
    default init method from pattern {chanel}_{YYYY/MM/DD/HH/mm/ss}.{fileformat}
    e.g: "channel_YYYY/MM/DD/HH/mm/ss.jpg"
    """

    def __init__(self, channel: str, datetime: Arrow, format: str = 'jpg',
                 date_str_format='YYYY/MM/DD/HH/mm/ss'):
        self.date_str = datetime.format(date_str_format)
        self.channel = channel
        self.format = format

    def __hash__(self):
        return hash(self.date_str) ^ hash(self.channel) ^ hash(self.format)

    def __str__(self):
        return "{}_{}.{}".format(self.channel, self.date_str, self.format)

    def __eq__(self, other):
        if isinstance(other, ImageId):
            return str(self) == str(other)
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, ImageId):
            return str(self) < str(other)
        return NotImplemented
